package flink.zanxiangnet.ad.monitoring;

import com.aliyun.odps.Instance;
import com.aliyun.odps.Odps;
import com.aliyun.odps.account.Account;
import com.aliyun.odps.account.AliyunAccount;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.task.SQLTask;
import com.tencent.ads.model.DailyReportsGetListStruct;
import com.tencent.ads.model.HourlyReportsGetListStruct;
import flink.zanxiangnet.ad.monitoring.clickhouse.sink.BatchSinkHour;
import flink.zanxiangnet.ad.monitoring.clickhouse.sink.BatchSinkMinute;
import flink.zanxiangnet.ad.monitoring.maxcompute.MaxComputeLog;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdDataOfDayDTO;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdStatOfDayODSDTO;
import flink.zanxiangnet.ad.monitoring.pojo.entity.*;
import flink.zanxiangnet.ad.monitoring.pojo.properties.ApplicationProperties;
import flink.zanxiangnet.ad.monitoring.pojo.dto.AdDataOfMinuteDTO;
import flink.zanxiangnet.ad.monitoring.process.AdMinuteDWDProcess;
import flink.zanxiangnet.ad.monitoring.process.CostHourDayProcess;
import flink.zanxiangnet.ad.monitoring.process.CostHourProcess;
import flink.zanxiangnet.ad.monitoring.process.CostMinuteProcess;
import flink.zanxiangnet.ad.monitoring.sink.TunnelBatchStreamSink;
import flink.zanxiangnet.ad.monitoring.stream.KeyedBatchStream;
import flink.zanxiangnet.ad.monitoring.trigger.AdMinuteODSStreamTrigger;
import flink.zanxiangnet.ad.monitoring.trigger.CostHourDMStreamTrigger;
import flink.zanxiangnet.ad.monitoring.trigger.CostMinuteDMStreamTrigger;
import flink.zanxiangnet.ad.monitoring.util.DateUtil;
import flink.zanxiangnet.ad.monitoring.util.JsonUtil;
import flink.zanxiangnet.ad.monitoring.kafka.KafkaComponent;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.flink.api.common.eventtime.*;
import org.apache.flink.api.common.state.MapState;
import org.apache.flink.api.common.state.MapStateDescriptor;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.connector.kafka.source.KafkaSource;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.DataStreamSource;
import org.apache.flink.streaming.api.datastream.SingleOutputStreamOperator;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.ProcessFunction;
import org.apache.flink.streaming.api.functions.windowing.ProcessWindowFunction;
import org.apache.flink.streaming.api.windowing.assigners.TumblingEventTimeWindows;
import org.apache.flink.streaming.api.windowing.time.Time;
import org.apache.flink.streaming.api.windowing.windows.GlobalWindow;
import org.apache.flink.util.Collector;
import org.apache.flink.util.OutputTag;
import org.springframework.beans.BeanUtils;

import java.time.Duration;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.*;
import java.util.stream.Collectors;

@Slf4j
public class AdStatJob {

    /**
     * 可能有数据的最早日期
     */
    private static final String OLDEST_DAY = "2019-01-01";

    public static void main(String[] args) throws Exception {
        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();

        // 加载配置文件到 flink的全局配置中
        Properties props = new Properties();
        props.load(AdStatJob.class.getResourceAsStream("/application.properties"));
        Configuration configuration = new Configuration();
        props.stringPropertyNames().forEach(key -> {
            String value = props.getProperty(key);
            configuration.setString(key.trim(), StringUtils.isBlank(value) ? "" : value.trim());
        });
        env.getConfig().setGlobalJobParameters(configuration);

        KafkaSource<String> adStreamOfMinuteSource = KafkaComponent.buildKafkaSource(props, KafkaComponent.KafkaTopic.adHourTopic, KafkaComponent.KafkaTopic.KafkaGroupId.adHourConsumerGroup);
        DataStreamSource<String> adStreamOfMinuteIn = env.fromSource(adStreamOfMinuteSource, WatermarkStrategy.noWatermarks(), "adHourSource_kafka");

        // 广告分钟数据（前 5分钟的广告消耗数据）
        final OutputTag<AdDataOfMinuteODS> adMinuteStreamTag = new OutputTag<AdDataOfMinuteODS>("adMinuteStream") {
        };
        // 广告小时数据（往前回滚 10天）
        final OutputTag<AdDataOfHourODS> adHourStreamTag = new OutputTag<AdDataOfHourODS>("adHourStream") {
        };

        // 对流进行映射，拆分（实时的分钟流和回滚的小时流）
        SingleOutputStreamOperator<AdDataOfMinuteODS> adODSStream = adStreamOfMinuteIn.filter(StringUtils::isNotBlank)
                .process(new ProcessFunction<String, AdDataOfMinuteODS>() {
                    @Override
                    public void processElement(String adDataStr, ProcessFunction<String, AdDataOfMinuteODS>.Context context, Collector<AdDataOfMinuteODS> collector) throws Exception {
                        AdDataOfMinuteDTO dto = JsonUtil.toObj(adDataStr, AdDataOfMinuteDTO.class);
                        // 指记录被创建的时间
                        long createTime = dto.getCreateTime();
                        // 指记录统计的时间（如果是实时统计的数据精确到分钟，回滚的历史数据精确到天）
                        long statTime = DateUtil.localDateTimeToMilli(DateUtil.milliToLocalDateTime(dto.getDataTime()).withSecond(0).withNano(0));
                        HourlyReportsGetListStruct struct = dto.getHourlyReportsGetListStruct();

                        if (createTime - statTime > (60 * 60 * 1000L)) {
                            AdDataOfHourODS adODS = new AdDataOfHourODS();
                            BeanUtils.copyProperties(struct, adODS);
                            adODS.setStatDay(DateUtil.formatLocalDate(DateUtil.milliToLocalDate(statTime)));
                            adODS.setHour(dto.getHourlyReportsGetListStruct().getHour().intValue());
                            adODS.setStatTime(DateUtil.localDateTimeToMilli(DateUtil.milliToLocalDateTime(statTime).withMinute(0)));
                            adODS.setAccountId(dto.getAccountId());
                            adODS.setAgencyAccountId(dto.getHourlyReportsGetListStruct().getAccountId());
                            adODS.setCreateTime(new Date(createTime));
                            adODS.removeNull();
                            context.output(adHourStreamTag, adODS);
                        } else {
                            AdDataOfMinuteODS adODS = new AdDataOfMinuteODS();
                            BeanUtils.copyProperties(struct, adODS);
                            adODS.setStatDay(DateUtil.formatLocalDate(DateUtil.milliToLocalDate(statTime)));
                            adODS.setHour(dto.getHourlyReportsGetListStruct().getHour().intValue());
                            adODS.setStatTime(statTime);
                            adODS.setAccountId(dto.getAccountId());
                            adODS.setAgencyAccountId(dto.getHourlyReportsGetListStruct().getAccountId());
                            adODS.setCreateTime(new Date(createTime));
                            adODS.removeNull();
                            context.output(adMinuteStreamTag, adODS);
                        }
                    }
                });

        // 分钟流
        DataStream<AdDataOfMinuteODS> adMinuteODSStream = adODSStream.getSideOutput(adMinuteStreamTag);
        // 写入原始表
        // adMinuteODSStream.addSink(new TunnelBatchSink<>(AdDataOfMinuteODS.class, 36000L, 64000L, 3));
        new KeyedBatchStream<>("adMinuteODSStream", adMinuteODSStream.keyBy(AdDataOfMinuteODS::getStatDay), 4000L, 2 * 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfMinuteODS.class))
                .name("sink_ad_minute_ods");

        SingleOutputStreamOperator<AdStatOfMinuteDWD> adMinuteDWDStream = adMinuteODSStream
                // 打水印，允许数据延迟 6分钟，同时指定时间流
                .assignTimestampsAndWatermarks(WatermarkStrategy.<AdDataOfMinuteODS>forBoundedOutOfOrderness(Duration.ofMinutes(6L))
                        .withTimestampAssigner((SerializableTimestampAssigner<AdDataOfMinuteODS>) (adODS, l) -> adODS.getStatTime()))
                .keyBy(AdDataOfMinuteODS::getAdId)
                // 开一个 5分钟的滚动窗口
                .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                .trigger(new AdMinuteODSStreamTrigger())
                .process(new AdMinuteDWDProcess());
        // .addSink(new TunnelBatchSink<>(AdStatOfMinuteDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adMinuteDWDStream", adMinuteDWDStream.keyBy(AdStatOfMinuteDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfMinuteDWD.class))
                .name("sink_ad_minute_dwd");

        //cost----分钟数据处理
        SingleOutputStreamOperator<CostMinuterDM> clickhouseMinuteDmStream =
                adMinuteDWDStream
                        .keyBy(AdStatOfMinuteDWD::getAdId)
                        .window(TumblingEventTimeWindows.of(Time.minutes(5L)))
                        .trigger(new CostMinuteDMStreamTrigger())
                        .process(new CostMinuteProcess())
                        .name("sink_ad_minute_dm_clickhouse");

        BatchSinkMinute batchSinkMinute = new BatchSinkMinute();
        clickhouseMinuteDmStream.addSink(batchSinkMinute);


        // 小时流（直接写到小时报表的 ods）
        DataStream<AdDataOfHourODS> adHourODSStream = adODSStream.getSideOutput(adHourStreamTag);
        // 写入原始表
        // adHourODSStream.addSink(new TunnelBatchSink<>(AdDataOfHourODS.class, 36000L, 64000L, 10));
        new KeyedBatchStream<>("adHourODSStream", adHourODSStream.keyBy(AdDataOfHourODS::getStatDay), 4000L, 5 * 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfHourODS.class))
                .name("sink_ad_hour_ods");

        SingleOutputStreamOperator<AdStatOfHourDWD> adHourDWDStream =
                adHourODSStream.keyBy(AdDataOfHourODS::getAdId)
                .countWindow(1).
                process(new ProcessWindowFunction<AdDataOfHourODS, AdStatOfHourDWD, Long, GlobalWindow>() {
                    private Odps odps;
                    // 上次查询的天数据
                    private ValueState<String> lastQueryDayState;
                    // 聚合的天的数据
                    private MapState<String, AdStatOfHourDWD> historyReduceState;

                    @Override
                    public void open(Configuration conf) {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));

                        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
                        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", Types.STRING, Types.POJO(AdStatOfHourDWD.class)));
                    }

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfHourODS, AdStatOfHourDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfHourODS> iterable, Collector<AdStatOfHourDWD> collector) throws Exception {
                        AdDataOfHourODS element = iterable.iterator().next();
                        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
                        LocalDateTime statTime = LocalDateTime.of(statDay, LocalTime.of(element.getHour(), 0, 0));
                        long now = System.currentTimeMillis();
                        LocalDate today = LocalDate.now();

                        String lastQueryDay = lastQueryDayState.value();
                        // 从 maxCompute拉取指定 广告的历史数据
                        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(today))) {
                            LocalDate endDay = today, beginDay = statDay.minusDays(60);
                            String sql = "SELECT * FROM ad_stat_of_hour_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginDay) + "\" AND stat_day <= \"" + DateUtil.formatLocalDate(endDay) + "\" AND ad_id = " + element.getAdId() + ";";
                            Instance instance = SQLTask.run(odps, sql);
                            System.out.println("212===>sql: " + sql + ", odps日志: " + odps.logview().generateLogView(instance, 7 * 24));
                            instance.waitForSuccess();
                            List<Record> records = SQLTask.getResult(instance);
                            Map<String, AdStatOfHourDWD> historyHourMap = records.stream()
                                    .map(AdStatOfHourDWD::byMaxCompute)
                                    .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue())
                                    .collect(Collectors.toMap(data -> data.getStatDay() + data.getHour(), data -> data, (val1, val2) -> val2));
                            historyReduceState.clear();
                            historyReduceState.putAll(historyHourMap);
                            lastQueryDayState.update(DateUtil.formatLocalDate(today));
                        }
                        AdStatOfHourDWD lastReduceData = null;
                        for (int i = 1; i < 60 * 24; i++) {
                            LocalDateTime time = statTime.minusHours(i);
                            lastReduceData = historyReduceState.get(DateUtil.formatLocalDate(time.toLocalDate()) + time.getHour());
                            if (lastReduceData != null) {
                                break;
                            }
                        }

                        AdStatOfHourDWD newStatData = AdStatOfHourDWD.reduce(lastReduceData, element, now);
                        collector.collect(newStatData);
                    }
                });
        //.addSink(new TunnelBatchSink<>(AdStatOfHourDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adHourDWDStream", adHourDWDStream.keyBy(AdStatOfHourDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfHourDWD.class))
                .name("sink_ad_hour_dwd");

        //小时数据
        SingleOutputStreamOperator<CostHourDM> clickhouseHourDmStream =
                adHourDWDStream
                        .keyBy(AdStatOfHourDWD::getAdId)
                        .countWindow(1)
                        .process(new CostHourProcess())
                        .name("sink_ad_hour_dm_clickhouse");

//        clickhouseHourDmStream.print();
        BatchSinkHour batchSinkhour = new BatchSinkHour();
        clickhouseHourDmStream.addSink(batchSinkhour);


        // ------------------------------------------------------- 处理广告的天数据 -----------------------------------------
        KafkaSource<String> adStreamOfDaySource = KafkaComponent.buildKafkaSource(props, KafkaComponent.KafkaTopic.adDayTopic, KafkaComponent.KafkaTopic.KafkaGroupId.adDayConsumerGroup);

        DataStreamSource<String> adStreamOfDayIn = env.fromSource(adStreamOfDaySource, WatermarkStrategy.noWatermarks(), "adDaySource_kafka");

        // 广告日数据。往前回滚 10天
        final OutputTag<AdDataOfDayODS> adDayStreamRollDayTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollDayTag") {
        };
        // 广告日数据。往前回滚 1年
        final OutputTag<AdDataOfDayODS> adDayStreamRollYearTag = new OutputTag<AdDataOfDayODS>("adDayStreamRollYearTag") {
        };
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStream = adStreamOfDayIn.filter(StringUtils::isNotBlank)
                .map(str -> {
                    AdDataOfDayDTO dto = JsonUtil.toObj(str, AdDataOfDayDTO.class);
                    Date createTime = new Date(dto.getCreateTime());
                    DailyReportsGetListStruct struct = dto.getDailyReportsGetListStruct();
                    AdDataOfDayODS adODS = new AdDataOfDayODS();
                    BeanUtils.copyProperties(struct, adODS);
                    adODS.setStatDay(struct.getDate());
                    adODS.setAccountId(dto.getAccountId());
                    adODS.setCampaignId(struct.getCampaignId());
                    adODS.setAgencyAccountId(struct.getAccountId());
                    adODS.setWechatAccountId(struct.getWechatAccountId());
                    adODS.setAdgroupId(struct.getAdgroupId());
                    adODS.setAdId(struct.getAdId());
                    adODS.setCreateTime(createTime);
                    adODS.removeNull();
                    return AdStatOfDayODSDTO.builder()
                            .startDate(dto.getStartDate())
                            .endDate(dto.getEndDate())
                            .adDataOfDayODS(adODS)
                            .build();
                });
        // adDayODSStream.map(AdStatOfDayODSDTO::getAdDataOfDayODS).addSink(new TunnelBatchSink<>(AdDataOfDayODS.class, 36000L, 6000L, 10));
        new KeyedBatchStream<>("adDayODSStream", adDayODSStream.map(AdStatOfDayODSDTO::getAdDataOfDayODS).keyBy(AdDataOfDayODS::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdDataOfDayODS.class))
                .name("sink_ad_day_ods");

        // 拆分流
        SingleOutputStreamOperator<AdStatOfDayODSDTO> adDayODSStreamSplit = adDayODSStream.process(new ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>() {
            @Override
            public void processElement(AdStatOfDayODSDTO adDataDTO, ProcessFunction<AdStatOfDayODSDTO, AdStatOfDayODSDTO>.Context context, Collector<AdStatOfDayODSDTO> collector) throws Exception {
                LocalDate startDate = adDataDTO.getStartDate();
                LocalDate endDate = adDataDTO.getEndDate();
                AdDataOfDayODS adODS = adDataDTO.getAdDataOfDayODS();
                if (DateUtil.intervalOfDays(startDate, endDate) > 31L) {
                    // 拉取时间间隔超过 1个月，账号回滚 365天的数据
                    context.output(adDayStreamRollYearTag, adODS);
                } else {
                    // 每日往前回滚 10天的数据
                    context.output(adDayStreamRollDayTag, adODS);
                }
            }
        });

        SingleOutputStreamOperator<AdStatOfDayDWD> adDayDWDStream = adDayODSStreamSplit.getSideOutput(adDayStreamRollDayTag)
                .keyBy(AdDataOfDayODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>() {
                    private Odps odps;
                    // 上次查询的时间
                    private ValueState<String> lastQueryDayState;
                    // 之前聚合的昨天的数据
                    private MapState<String, AdStatOfDayDWD> historyReduceState;

                    @Override
                    public void open(Configuration conf) {
                        Map<String, String> params = getRuntimeContext()
                                .getExecutionConfig()
                                .getGlobalJobParameters()
                                .toMap();
                        Account account = new AliyunAccount(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ID),
                                params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_KEY));
                        odps = new Odps(account);
                        odps.getRestClient().setRetryLogger(new MaxComputeLog());
                        odps.setEndpoint(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_ENDPOINT));
                        odps.setDefaultProject(params.get(ApplicationProperties.MAX_COMPUTE_ACCOUNT_PROJECT_NAME));

                        lastQueryDayState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastQueryDayState", String.class));
                        historyReduceState = getRuntimeContext().getMapState(new MapStateDescriptor<>("historyReduceState", String.class, AdStatOfDayDWD.class));
                    }

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>.Context context, Iterable<AdDataOfDayODS> iterable, Collector<AdStatOfDayDWD> collector) throws Exception {
                        AdDataOfDayODS element = iterable.iterator().next();
                        LocalDate statDay = DateUtil.parseLocalDate(element.getStatDay());
                        long now = System.currentTimeMillis();

                        String lastQueryDay = lastQueryDayState.value();
                        // 从 maxCompute查找广告的历史数据
                        if (lastQueryDay == null || !lastQueryDay.equals(DateUtil.formatLocalDate(LocalDate.now()))) {
                            LocalDate endTime = LocalDate.now(), beginTime = statDay.minusDays(60);
                            String sql = "SELECT * FROM ad_stat_of_day_dwd WHERE stat_day >= \"" + DateUtil.formatLocalDate(beginTime) + "\" AND stat_day <= \"" + DateUtil.formatLocalDate(endTime) + "\" AND ad_id = " + element.getAdId() + ";";
                            Instance instance = SQLTask.run(odps, sql);
                            System.out.println("337===>sql: " + sql + ", odps日志: " + odps.logview().generateLogView(instance, 7 * 24));
                            instance.waitForSuccess();
                            List<Record> records = SQLTask.getResult(instance);
                            Map<String, AdStatOfDayDWD> historyData = records.stream()
                                    .map(AdStatOfDayDWD::byMaxCompute)
                                    .sorted((o1, o2) -> new Long(o1.getCreateTime().getTime() - o2.getCreateTime().getTime()).intValue())
                                    .collect(Collectors.toMap(AdStatOfDayDWD::getStatDay, data -> data, (val1, val2) -> val2));
                            historyReduceState.clear();
                            historyReduceState.putAll(historyData);
                            lastQueryDayState.update(DateUtil.formatLocalDate(LocalDate.now()));
                        }
                        AdStatOfDayDWD lastReduceData = null;
                        for (int i = 1; i <= 60; i++) {
                            lastReduceData = historyReduceState.get(DateUtil.formatLocalDate(statDay.minusDays(i)));
                            if (lastReduceData != null) {
                                break;
                            }
                        }
                        AdStatOfDayDWD newStatData = AdStatOfDayDWD.reduce(lastReduceData, element, now);
                        historyReduceState.put(DateUtil.formatLocalDate(statDay), newStatData);
                        collector.collect(newStatData);
                    }
                });
        //.addSink(new TunnelBatchSink<>(AdStatOfDayDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adDayDWDStream", adDayDWDStream.keyBy(AdStatOfDayDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfDayDWD.class))
                .name("sink_ad_day_dwd");

        //小时数据-----天
        SingleOutputStreamOperator<CostHourDM> clickhouseHourDayDmStream =
                adDayDWDStream
                        .keyBy(AdStatOfDayDWD::getAdId)
                        .countWindow(1)
                        .process(new CostHourDayProcess())
                        .name("sink_ad_hour_day_dm_clickhouse");

//        clickhouseHourDayDmStream.print();
        BatchSinkHour batchSinkHourDay = new BatchSinkHour();
        clickhouseHourDayDmStream.addSink(batchSinkHourDay);


        SingleOutputStreamOperator<AdStatOfDayDWD> adDayDWDYearStream = adDayODSStreamSplit.getSideOutput(adDayStreamRollYearTag)
                .keyBy(AdDataOfDayODS::getAdId)
                .countWindow(1).process(new ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>() {
                    // 上次聚合的结果
                    private ValueState<AdStatOfDayDWD> lastReduceState;

                    public void open(Configuration conf) {
                        lastReduceState = getRuntimeContext().getState(new ValueStateDescriptor<>("lastReduceState", AdStatOfDayDWD.class));
                    }

                    @Override
                    public void process(Long elementsCount, ProcessWindowFunction<AdDataOfDayODS, AdStatOfDayDWD, Long, GlobalWindow>.Context context,
                                        Iterable<AdDataOfDayODS> elements, Collector<AdStatOfDayDWD> out) throws Exception {
                        AdDataOfDayODS element = elements.iterator().next();
                        AdStatOfDayDWD newStatDWD = AdStatOfDayDWD.reduce(lastReduceState.value(), element, System.currentTimeMillis());
                        lastReduceState.update(newStatDWD);
                        out.collect(newStatDWD);
                    }
                });
        //.addSink(new TunnelBatchSink<>(AdStatOfDayDWD.class, 30000L, 365L, 6));
        new KeyedBatchStream<>("adDayDWDYearStream", adDayDWDYearStream.keyBy(AdStatOfDayDWD::getStatDay), 4000L, 60 * 1000L)
                .toBatch()
                .addSink(new TunnelBatchStreamSink<>(AdStatOfDayDWD.class))
                .name("sink_ad_year_dwd");
        //小时数据----年
        SingleOutputStreamOperator<CostHourDM> clickhouseHourYearDmStream =
                adDayDWDYearStream
                        .keyBy(AdStatOfDayDWD::getAdId)
                        .countWindow(1)
                        .process(new CostHourDayProcess())
                        .name("sink_ad_hour_year_dm_clickhouse");

//        clickhouseHourYearDmStream.print();
        BatchSinkHour batchSinkHourYear = new BatchSinkHour();
        clickhouseHourYearDmStream.addSink(batchSinkHourYear);

        //TODO:jobname----后面应该带个时间以此来区分log以及job
        env.execute("job_ad");
    }
}
