package flink.zanxiangnet.ad.monitoring.clickhouse.bean;

import flink.zanxiangnet.ad.monitoring.clickhouse.bean.annotation.ClickhouseColumn;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Data;
import lombok.NoArgsConstructor;
import org.apache.commons.lang3.StringUtils;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class BeanUtil {
    private static final Pattern humpPattern = Pattern.compile("[A-Z]");
    private static final Pattern linePattern = Pattern.compile("_(\\w)");

    /**
     * 解析 javabean的字段
     *
     * @param clazz
     * @return
     */
    public static List<FieldInfo> parseBeanField(Class<?> clazz) {
        List<Field> fieldList = new ArrayList<>();
        Class<?> tempClazz = clazz;
        while (tempClazz != null) {
            Field[] fields = tempClazz.getDeclaredFields();
            if (fields.length != 0) {
                fieldList.addAll(Arrays.asList(fields));
            }
            tempClazz = tempClazz.getSuperclass();
        }
        if (fieldList.size() == 0) {
            throw new RuntimeException("Failed parsed class[" + clazz.getName() + "]. No Field!");
        }
        List<FieldInfo> fieldInfoList = new ArrayList<>(fieldList.size());
        for (Field field : fieldList) {
            if (Modifier.isFinal(field.getModifiers()) || Modifier.isStatic(field.getModifiers())) {
                continue;
            }
            ClickhouseColumn annotation = field.getAnnotation(ClickhouseColumn.class);
            FieldInfo fieldInfo = new FieldInfo();
            fieldInfo.setField(field);
            fieldInfo.setFieldName(field.getName());
            try {
                fieldInfo.setGetMethod(methodOfGet(clazz, field));
                fieldInfo.setSetMethod(methodOfSet(clazz, field));
            } catch (NoSuchMethodException e) {
                throw new RuntimeException("Failed parsed class[" + clazz.getName() + "]. The get or set method of [" + field.getName() + "] was not found!");
            }
            if (annotation != null) {
                if (annotation.ignore()) {
                    continue;
                }
                if (annotation.isPartitioned()) {
                    fieldInfo.setUsePartitioned(true);
                }
                if (StringUtils.isNotBlank(annotation.value())) {
                    fieldInfo.setColumnName(annotation.value());
                } else {
                    fieldInfo.setColumnName(humpToLine(fieldInfo.getFieldName()));
                }
            } else {
                fieldInfo.setColumnName(humpToLine(fieldInfo.getFieldName()));
            }
            fieldInfoList.add(fieldInfo);
        }
        return fieldInfoList;
    }

    /**
     * 驼峰转下划线
     *
     * @param str
     * @return
     */
    public static String humpToLine(String str) {
        Matcher matcher = humpPattern.matcher(str);
        StringBuffer sb = new StringBuffer();
        while (matcher.find()) {
            matcher.appendReplacement(sb, "_" + matcher.group(0).toLowerCase());
        }
        matcher.appendTail(sb);
        return sb.toString();
    }

    /**
     * 下划线转驼峰
     *
     * @param str
     * @return
     */
    public static String lineToHump(String str) {
        str = str.toLowerCase();
        Matcher matcher = linePattern.matcher(str);
        StringBuffer sb = new StringBuffer();
        while (matcher.find()) {
            matcher.appendReplacement(sb, matcher.group(1).toUpperCase());
        }
        matcher.appendTail(sb);
        return sb.toString();
    }

    public static <T> Method methodOfGet(Class<T> clazz, Field field) throws NoSuchMethodException {
        String fieldName = field.getName();
        String methodName = "get" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
        return clazz.getMethod(methodName);
    }

    public static <T> Method methodOfSet(Class<T> clazz, Field field) throws NoSuchMethodException {
        String fieldName = field.getName();
        String methodName = "set" + fieldName.substring(0, 1).toUpperCase() + fieldName.substring(1);
        return clazz.getMethod(methodName, field.getType());
    }

    @Data
    @NoArgsConstructor
    @AllArgsConstructor
    @Builder
    public static class FieldInfo {
        /**
         * 是否用于分
         */
        private boolean usePartitioned = false;
        /**
         * 字段名
         */
        private String fieldName;
        /**
         * 数据库中的列名
         */
        private String columnName;
        /**
         * 字段
         */
        private Field field;
        /**
         * 字段的 get方法
         */
        private Method getMethod;

        /**
         * 字段的 set方法
         */
        private Method setMethod;
    }
}
