package flink.zanxiangnet.ad.monitoring.clickhouse.sink;

import com.aliyun.odps.PartitionSpec;
import com.aliyun.odps.data.Record;
import com.aliyun.odps.data.RecordWriter;
import com.aliyun.odps.tunnel.TableTunnel;
import com.aliyun.odps.tunnel.TunnelException;
import flink.zanxiangnet.ad.monitoring.clickhouse.bean.BeanUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.CollectionUtils;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;

/**
 * 实际执行 MaxCompute写入逻辑
 */
public class TunnelBatchWriter<T> implements Runnable {
    private static final Logger log = LoggerFactory.getLogger(TunnelBatchWriter.class);

    private final TableTunnel tunnel;
    private final String projectName;
    private final String tableName;
    private final PartitionSpec part;
    private final List<T> dataList;
    private final List<BeanUtil.FieldInfo> fieldInfoList;

    public TunnelBatchWriter(TableTunnel tunnel,
                             String projectName,
                             String tableName,
                             PartitionSpec part,
                             List<T> dataList,
                             List<BeanUtil.FieldInfo> fieldInfoList
    ) {
        this.tunnel = tunnel;
        this.projectName = projectName;
        this.tableName = tableName;
        this.part = part;
        this.dataList = dataList;
        this.fieldInfoList = fieldInfoList;
    }

    @Override
    public void run() {
        if (CollectionUtils.isEmpty(dataList)) {
            return;
        }
        int retry = 0;
        do {
            try {
                if (retry > 0) {
                    log.error("重试：" + retry);
                }
                // 创建 session大概要 200多毫秒
                TableTunnel.UploadSession uploadSession = part == null ? tunnel.createUploadSession(projectName, tableName)
                        : tunnel.createUploadSession(projectName, tableName, part);
                long blockId = 0;
                RecordWriter writer = uploadSession.openRecordWriter(blockId);
                for (T t : dataList) {
                    Record record = uploadSession.newRecord();
                    for (BeanUtil.FieldInfo fieldInfo : fieldInfoList) {
                        if (fieldInfo.isUsePartitioned()) {
                            // 分区字段不在这里设值
                            continue;
                        }
                        Object obj = fieldInfo.getGetMethod().invoke(t);
                        record.set(fieldInfo.getColumnName(), obj);
                    }
                    writer.write(record);
                }
                writer.close();
                uploadSession.commit(new Long[]{blockId});
                break;
            } catch (IOException | InvocationTargetException | IllegalAccessException | TunnelException e) {
                log.error(e.getMessage(), e);
                if (++retry >= 3) {
                    log.error("重试失败");
                    throw new RuntimeException("Failed writer to MaxCompute!!! Error msg: " + e.getMessage());
                }
            } catch (Exception e) {
                if (++retry >= 3) {
                    log.error("重试失败");
                    log.error(e.getMessage(), e);
                    throw new RuntimeException("Failed writer to MaxCompute!!! Error msg: " + e.getMessage());
                }
            }
        } while (true);
    }
}